from pathlib import Path
import boto3
import os
from configparser import ConfigParser
from requests_aws4auth import AWS4Auth


class credentialsFile:
    
    def __init__(self, credential_path=None, config_path=None, profile_name=None):
        ''' Establish environment variables for boto3 authentication '''
        if profile_name is None:
            profile_name = 'test'
        if os.getenv(profile_name.upper() + '_AWS_PROFILE') is None:
            self.AWS_SHARED_CREDENTIALS_FILE = str(Path(credential_path).expanduser())
            self.AWS_CONFIG_FILE = str(Path(config_path).expanduser())
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = self.AWS_SHARED_CREDENTIALS_FILE
            os.environ['AWS_CONFIG_FILE'] = self.AWS_CONFIG_FILE
            self.credentials = ConfigParser()
            self.credentials.read(self.AWS_SHARED_CREDENTIALS_FILE)
            config = ConfigParser()
            config.read(self.AWS_CONFIG_FILE)
            os.environ['AWS_PROFILE'] = profile_name
            os.environ['AWS_ACCESS_KEY_ID'] = self.credentials[profile_name]['aws_access_key_id']
            os.environ['AWS_SECRET_ACCESS_KEY'] = self.credentials[profile_name]['aws_secret_access_key']
            os.environ['AWS_DEFAULT_REGION'] = config['profile ' + profile_name]['region']
        else:
            os.environ['AWS_ACCESS_KEY_ID'] = os.getenv(profile_name.upper() + '_AWS_ACCESS_KEY_ID', 'test')
            os.environ['AWS_SECRET_ACCESS_KEY'] = os.getenv(
                profile_name.upper() + '_AWS_SECRET_ACCESS_KEY', 'test')
            os.environ['AWS_DEFAULT_REGION'] = os.getenv(
                profile_name.upper() + '_AWS_DEFAULT_REGION', 'test')


class establishSession(credentialsFile):
    ''' Establish AWS service sessions '''
    
    def __init__(self, credential_path=None, config_path=None, profile_name=None,
                service_name=None):
        super().__init__(credential_path=credential_path, config_path=config_path,
                        profile_name=profile_name)
        self.session = boto3.Session(aws_access_key_id=os.getenv('AWS_ACCESS_KEY_ID'),
                                    aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY'))
        self.access_key = self.session.get_credentials().access_key
        self.secret_key = self.session.get_credentials().secret_key
        self.region = self.session.region_name
        self.awsauth = AWS4Auth(self.access_key, self.secret_key,
                                self.region, service_name)


class establishES(establishSession):
    ''' Establish AWS elasticsearch session '''

    def __init__(self, credential_path=None, config_path=None, profile_name=None,
                elastic_path=None, service_name='es'):
        super().__init__(credential_path=credential_path, config_path=config_path,
                        profile_name=profile_name, service_name=service_name)
        self.client = self.session.client(service_name)
        self._test = 'test'
        if elastic_path is not None:
            self.es_config = str(Path(elastic_path).expanduser())
            config = ConfigParser()
            config.read(self.es_config)
            self.es_endpoint = config['elasticsearch-endpoint']['address']
        else:
            self.es_endpoint = os.getenv("FTX_ENDPOINT")


class establishS3(establishSession):
    ''' Establish connection with AWS S3 '''

    def __init__(self, credential_path=None, config_path=None, profile_name=None, service_name='s3'):
        super().__init__(credential_path=credential_path, config_path=config_path, profile_name=profile_name,
                        service_name=service_name)
        self.resource = self.session.resource(service_name)
        self.client = self.resource.meta.client


class establishQuilt(credentialsFile):
    ''' Create authentication files for quilt3 based on aws_credentials '''

    def __init__(self, credential_path=None, config_path=None, profile_name=None):
        super().__init__(credential_path=credential_path, config_path=config_path, 
                        profile_name=profile_name)
        if profile_name is None:
            profile_name = 'quilt-karrlab'
        if os.getenv(profile_name.upper() + '_AWS_PROFILE') is None:
            os.environ['EXPIRY_TIME'] = '2029-09-30T20:32:07+00:00'



import json
from pathlib import Path, PurePath


class IndexUtil:
    """Make index setting json file.
    """

    def __init__(self, filter_dir=None, analyzer_dir=None, mapping_properties_dir=None):
        """Initialize
        
        Args:
            filter_dir (:obj:`str`, optional): filter file directory. Defaults to None.
            analyzer_dir (:obj:`str`, optional): analyzer file directory. Defaults to None.
            mapping_properties_dir (:obj:`str`, optional): mapping properties file directory. Defaults to None.
        """
        self.filter_dir = filter_dir
        self.analyzer_dir = analyzer_dir
        self.mapping_properties_dir = mapping_properties_dir

    def read_file(self, _dir):
        """Read in json file.
        
        Args:
            _dir (:obj:`str`): directory of the json file.

        Return:
            (:obj:`dict`)
        """
        full_dir = str(Path(_dir).expanduser())
        with open(full_dir, 'r') as f:
            return json.load(f)

    def combine_files(self, **kwargs):
        """Combine various settings for index to form a coherent description.
        (https://www.elastic.co/guide/en/elasticsearch/reference/current/search-analyzer.html)

        Args:
            _filter (:obj:`bool`): whether to include filter info.
            analyzer (:obj:`bool`): whether to include analyzer info.
            mappings (:obj:`bool`): whether to include mappings info.

        Return:
            (:obj:`dict`)
        """
        _filter = kwargs.get('_filter')
        analyzer = kwargs.get('analyzer')
        mappings = kwargs.get('mappings')

        result = {}

        if _filter:
            _filter_data = self.read_file(self.filter_dir)
        if analyzer:
            analyzer_data = self.read_file(self.analyzer_dir)
        if mappings:
            mappings_data = self.read_file(self.mapping_properties_dir)

        result['settings'] = {'analysis': {**_filter_data, **analyzer_data}}
        result['mappings'] = mappings_data

        return result
        


""" karr_lab_aws_manager command line interface

:Author: Name <email>
:Date: 2019-9-16
:Copyright: 2019, Karr Lab
:License: MIT
"""

import cement
import tempfile
import karr_lab_aws_manager
import karr_lab_aws_manager.core
from karr_lab_aws_manager.elasticsearch_kl import util as es_util
from karr_lab_aws_manager.quilt3_kl import util as quilt_util
import json
import pprint


class BaseController(cement.Controller):
    """ Base controller for command line application """

    class Meta:
        label = 'base'
        description = "karr_lab_aws_manager"
        arguments = [
            (['-v', '--version'], dict(action='version', version=karr_lab_aws_manager.__version__)),
        ]

    @cement.ex(help='command_1 description')
    def cmd1(self):
        """ command_1 description """
        print('command_1 output')

    @cement.ex(help='command_2 description')
    def cmd2(self):
        """ command_2 description """
        print('command_2 output')

    @cement.ex(hide=True)
    def _default(self):
        self._parser.print_help()


class Command3WithArgumentsController(cement.Controller):
    """ Command3 description """

    class Meta:
        label = 'command-3'
        description = 'Command3 description'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['arg_1'], dict(
                type=str, help='Description of arg_1')),
            (['arg_2'], dict(
                type=str, help='Description of arg_2')),
            (['--opt-arg-3'], dict(
                type=str, default='default value of opt-arg-1', help='Description of opt-arg-3')),
            (['--opt-arg-4'], dict(
                type=float, default=float('nan'), help='Description of opt-arg-4')),
        ]

    @cement.ex(hide=True)
    def _default(self):
        args = self.app.pargs
        args.arg_1
        args.arg_2
        args.opt_arg_3
        args.opt_arg_4


class EsDeleteIdx(cement.Controller):
    """Karrlab elasticsearch delete index. """

    class Meta:
        label = 'es-del-idx'
        description = 'Delete index in ES'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['index'], dict(
                type=str, help='Name of index in es')),
            (['--id'], dict(
                type=str, help='ID of the doc in index')),
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Delete elasticsearch index

            Args:
                index (:obj:`str`): name of index in es
                _id (:obj:`int`): id of the doc in index (optional)
        '''
        args = self.app.pargs
        es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).delete_index(args.index, _id=args.id)


class EsCheckSvr(cement.Controller):
    """Karrlab elasticsearch delete index. """

    class Meta:
        label = 'es-check-server'
        description = 'Check ES server health status'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Delete elasticsearch index

            Args:
                index (:obj:`str`): name of index in es
                _id (:obj:`int`): id of the doc in index (optional)
        '''
        args = self.app.pargs
        r = es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).index_health_status()
        print(r.content.decode('utf-8'))


class EsGetIdxMapping(cement.Controller):
    """Karrlab elasticsearch delete index. """

    class Meta:
        label = 'es-get-mapping'
        description = 'Check ES server health status'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['--index'], dict(
                type=str, default='.kibana_1', help='Comma separated names of indices in es')),
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Delete elasticsearch index

            Args:
                index (:obj:`str`): name of index in es
                _id (:obj:`int`): id of the doc in index (optional)
        '''
        args = self.app.pargs
        r = es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).get_index_mapping(index=args.index)
        content = json.loads(r.content.decode('utf-8'))        
        pprint.pprint(content)

class EsBulkUpload(cement.Controller):
    """ Karrlab elasticsearch bulk upload cli """

    class Meta:
        label = 'es-bulk-upload'
        description = 'Bulk loading data into karrlab hosted elasticsearch service'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['cursor'], dict(
                type=str, help='Pymongo.Cursor/directory to files to be loaded')),
            (['id'], dict(
                type=str, help='Key in mongo collection for identification')),
            (['--count'], dict(
                type=int, default=None, help='Cursor/file size')),
            (['--index'], dict(
                type=str, default='test', help='Name of index in es')),
            (['--bulk_size', '-bz'], dict(
                type=int, default=100, help='Name of unique key to be used as index for es')),
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables')),
            (['--headers'], dict(
                type=dict, default={ "Content-Type": "application/json" },
                help='Http header'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Load data into elasticsearch service

            Args:
                count (:obj:`int`): cursor size
                cursor (:obj:`pymongo.Cursor` or :obj:`iter`): documents to be PUT/POST to es
                index (:obj:`str`): name of unique key to be used as index for es
                bulk_size (:obj:`int`): number of documents in one PUT
                headers (:obj:`dict`): http header
                _id (:obj:`str`): key in mogno collection for identification

            Returns:
                (:obj:`set`): set of status codes
        '''
        args = self.app.pargs
        es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).data_to_es_bulk(
                       args.cursor, index=args.index, count=args.count, bulk_size=args.bulk_size,
                       _id=args.id, headers=args.headers)


class EsSetIdx(cement.Controller):
    """Karrlab elasticsearch settings for index. """

    class Meta:
        label = 'es-set-idx'
        description = 'Index settings in ES'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['index'], dict(
                type=str, help='Name of index in es')),
            (['replica_count'], dict(
                type=int, help='Number of replicas')),
            (['--shard_count', '-sc'], dict(
                type=int, default=1,
                help='Number of primary shards contained in the es cluster')
            ),
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables')),
            (['--headers'], dict(
                type=dict, default={ "Content-Type": "application/json" },
                help='Http header'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        """Setting index's shard and replica number in es cluster
        
        Args:
            index (str): name of index to be set
            number_of_replicas (int): number of replica shards to be used for the index
            number_of_shards (int): number of primary shards contained in the es cluster
            headers (dict): http request content header description

        Returns:
            (HTTPResponse): http response
        """
        args = self.app.pargs
        es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).index_settings(
                           args.index, args.replica_count, number_of_shards=args.shard_count, headers=args.headers
                       )


class EsSetAnalysis(cement.Controller):
    """Karrlab elasticsearch settings for index. """

    class Meta:
        label = 'es-set-analysis'
        description = 'Index settings in ES'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['index'], dict(
                type=str, help='Name of index in es')),
            (['filter_dir'], dict(
                type=str, help='Directory of filter file')),
            (['analyzer_dir'], dict(
                type=str, help='Directory of analyzer file')
            ),
            (['--profile_name', '-pn'], dict(
                type=str, default='es-poweruser',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            ),
            (['--elastic_path', '-ep'], dict(
                type=str, default='~/.wc/third_party/elasticsearch.ini',
                help='Directory for file containing aws elasticsearch service variables')),
            (['--headers'], dict(
                type=dict, default={ "Content-Type": "application/json" },
                help='Http header'))
        ]

    @cement.ex(hide=True)
    def _default(self):
        """Setting index's shard and replica number in es cluster
        
        Args:
            index (str): name of index to be set
            number_of_replicas (int): number of replica shards to be used for the index
            number_of_shards (int): number of primary shards contained in the es cluster
            headers (dict): http request content header description

        Returns:
            (HTTPResponse): http response
        """
        args = self.app.pargs
        es_util.EsUtil(profile_name=args.profile_name, credential_path=args.credential_path,
                       config_path=args.config_path, elastic_path=args.elastic_path).update_index_analysis(
                           args.index, args.filter_dir, args.analyzer_dir)


class QuiltAddToPackage(cement.Controller):
    """Karrlab add files to quilt3 package"""

    class Meta:
        label = 'quilt-add2-package'
        description = 'Used for uploading datanator package to quilt3'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['package_name'], dict(
                type=str, help='Package to be manipulated')),
            (['destination'], dict(
                type=str, help='destination in package')),
            (['source'], dict(
                type=str, help='Sources to be added to package')),
            (['--meta', '-m'], dict(
                type=str, default="{'test': 'test_meta'}",
                help='Number of primary shards contained in the es cluster')
            ),
            (['--default_remote_registry', '-drr'], dict(
                type=str, default='s3://karrlab',
                help='AWS profile to use for authentication')),
            (['--profile_name', '-pn'], dict(
                type=str, default='quilt-karrlab',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            )
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Specifically used for uploading datanator package to quilt3
        
            Args:
                source (:obj:`list` of :obj:`str`): sources to be added to package,
                                                      directories must end with '/'
                destination (:obj:`list` of :obj:`str` ): package(s) to be manipulated,
                                                            directories must end with '/'
                meta (:obj:`list` of :obj:`dict`): package meta

            Returns:
        '''
        args = self.app.pargs
        base = tempfile.mkdtemp()
        quilt_util.QuiltUtil(profile_name=args.profile_name, default_remote_registry=args.default_remote_registry,
                            aws_path=args.credential_path, cache_dir=base, config_path=args.config_path).add_to_package(
                            args.package_name, destination=args.destination, source=args.source, meta=args.meta)


class QuiltPushToS3(cement.Controller):
    """Karrlab push local quilt package to remote registry"""

    class Meta:
        label = 'quilt-push2-s3'
        description = 'Used for uploading datanator package to quilt3 s3 bucket'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['package_name'], dict(
                type=str, help='name of package in "username/packagename" format')),
            (['--registry', '-r'], dict(
                type=str, default='s3://karrlab',
                help='S3 URI where package should be pushed')),
            (['--destination', '-d'], dict(
                type=str, default='s3://karrlab',
                help='file landing destination in remote registry')
            ),
            (['--message', '-m'], dict(
                type=str, default='package upload',
                help='commit message')
            ),
            (['--default_remote_registry', '-drr'], dict(
                type=str, default='s3://karrlab',
                help='AWS profile to use for authentication')),
            (['--profile_name', '-pn'], dict(
                type=str, default='quilt-karrlab',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            )
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Push local package to remote registry

            Args:
                package_name (:obj:`str`): name of package in "username/packagename" format
                registry (:obj:`str`): S3 URI where package should be pushed
                destination (:obj:`str`): file landing destination in remote registry
                message (:obj:`str`): commit message
        '''
        args = self.app.pargs
        base = tempfile.mkdtemp()
        quilt_util.QuiltUtil(profile_name=args.profile_name, default_remote_registry=args.default_remote_registry,
                            aws_path=args.credential_path, cache_dir=base, config_path=args.config_path).push_to_remote(
                            args.package_name, registry=args.registry, destination=args.destination, message=args.message)


class QuiltRmPkg(cement.Controller):
    """Karrlab remove exsting quilt3 package. """

    class Meta:
        label = 'quilt-rm-pkg'
        description = 'Used for deleting package'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['package_name'], dict(
                type=str, help='name of package in "username/packagename" format')),
            (['--registry', '-r'], dict(
                type=str, default='s3://karrlab',
                help='S3 URI where package should be pushed')),
            (['--delete_from_s3', '-d'], dict(
                type=bool, default=True,
                help='Whether to delete data in the registry')
            ),
            (['--default_remote_registry', '-drr'], dict(
                type=str, default='s3://karrlab',
                help='AWS profile to use for authentication')),
            (['--profile_name', '-pn'], dict(
                type=str, default='quilt-karrlab',
                help='AWS profile to use for authentication')),
            (['--credential_path', '-cr'], dict(
                type=str, default='~/.wc/third_party/aws_credentials',
                help='Directory for aws credentials file')),
            (['--config_path', '-cp'], dict(
                type=str, default='~/.wc/third_party/aws_config',
                help='Directory for aws config file')
            )
        ]

    @cement.ex(hide=True)
    def _default(self):
        ''' Push local package to remote registry

            Args:
                package_name (:obj:`str`): name of package in "username/packagename" format
                registry (:obj:`str`): S3 URI where package should be pushed
                destination (:obj:`str`): file landing destination in remote registry
                message (:obj:`str`): commit message
        '''
        args = self.app.pargs
        base = tempfile.mkdtemp()
        quilt_util.QuiltUtil(profile_name=args.profile_name, default_remote_registry=args.default_remote_registry,
                            aws_path=args.credential_path, cache_dir=base, config_path=args.config_path).delete_package(
                            args.package_name, registry=args.registry, delete_from_s3=args.delete_from_s3)


class App(cement.App):
    """ Command line application """
    class Meta:
        label = 'karr_lab_aws_manager'
        base_controller = 'base'
        handlers = [
            BaseController,
            Command3WithArgumentsController,
            EsBulkUpload,
            EsCheckSvr,
            EsGetIdxMapping,
            EsDeleteIdx,
            EsSetIdx,
            EsSetAnalysis,
            QuiltAddToPackage,
            QuiltPushToS3,
            QuiltRmPkg
        ]

def main():
    with App() as app:
        app.run()

if __name__=='__main__':
    main()


from karr_lab_aws_manager.config import config
from pathlib import Path


class S3Util(config.establishS3):

    def __init__(self, profile_name=None, credential_path=None, config_path=None):
        ''' Interacting with aws s3 buckets
        '''
        super().__init__(profile_name=profile_name, credential_path=credential_path, config_path=config_path)


    def download_dir(self, dist, bucket, local='/tmp'):
        ''' Download all contents in a directory in s3 bucket iteratively
        
            Args:
                dist (:obj:`str`): s3 directory key representation
                bucket (:obj:`str`): name of bucket
                local (:obj:`str`): local directory to store downloaded content
        '''
        paginator = self.client.get_paginator('list_objects')
        for result in paginator.paginate(Bucket=bucket, Delimiter='/', Prefix=dist):
            if result.get('CommonPrefixes') is not None:
                for subdir in result.get('CommonPrefixes'):
                    self.download_dir(subdir.get('Prefix'), bucket, local=local)
            for file in result.get('Contents', []):
                dest_pathname = Path(local, file.get('Key'))
                if file.get('Key').endswith('/'):
                    dest_pathname.mkdir(parents=True, exist_ok=True)
                else:
                    Path(dest_pathname.parent).mkdir(parents=True, exist_ok=True)
                    dest_pathname.touch()
                self.resource.meta.client.download_file(bucket, file.get('Key'), str(dest_pathname))

    def del_dir(self, bucket_name, prefix):
        """delete 'directory' in s3 bucket
        
        Args:
            bucket_name (:obj:`str`): name of bucket
            prefix (:obj:`str`): name of 'directory'
        """
        bucket = self.resource.Bucket(bucket_name)
        bucket.objects.filter(Prefix=prefix).delete()


""" karr_lab_aws_manager

:Author: Name <email>
:Date: 2019-9-16
:Copyright: 2019, Karr Lab
:License: MIT
"""


class ExampleClass(object):
    """ Descipton of ExampleClass

    Attributes:
        attr_1 (:obj:`type of attr_1`): description of attr_1
        attr_2 (:obj:`type of attr_2`): description of attr_2
        ...
    """

    def __init__(self, arg_1, arg_2, kwarg_1=None, kwarg_2=None):
        """
        Args:
            arg_1 (:obj:`type of arg_1`): description of arg_1
            arg_2 (:obj:`type of arg_2`): description of arg_2
            kwarg_1 (:obj:`type of kwarg_1`, optional): description of kwarg_1
            kwarg_2 (:obj:`type of kwarg_2`, optional): description of kwarg_2
            ...
        """
        self.attr_1 = arg_1
        self.attr_2 = arg_2

    def method_1(self, arg_1, arg_2, kwarg_1=None, kwarg_2=None):
        """ Description of method_1

        Args:
            arg_1 (:obj:`type of arg_1`): description of arg_1
            arg_2 (:obj:`type of arg_2`): description of arg_2
            kwarg_1 (:obj:`type of kwarg_1`, optional): description of kwarg_1
            kwarg_2 (:obj:`type of kwarg_2`, optional): description of kwarg_2
            ...

        Returns:
            :obj:`type of return value`: description of return value

        Raises:
            :obj:`type of raised exception(s)`: description of raised exceptions
        """
        pass


import quilt3
from karr_lab_aws_manager.config import config
from karr_lab_aws_manager.s3 import util as s3_util
import json
import os
from configparser import ConfigParser
from pathlib import Path, PurePath
import tempfile
from boto3.s3.transfer import TransferConfig
import botocore


class QuiltUtil(config.establishQuilt):

    def __init__(self, profile_name=None, default_remote_registry=None,
                aws_path=None, cache_dir=None, config_path=None):
        ''' Handle Quilt authentication file creation without having to use quilt3.login()

            Args:
                aws_path (:obj:`str`): directory in which aws credentials file resides
                base_path (:obj:`str`): directory to store quilt3 credentials generated from aws credentials
                profile_name (:obj:`str`): AWS credentials profile name for quilt
                default_remote_registry (:obj:`str`): default remote registry to store quilt package
                cache_dir (:obj:`str`): default directory to store data
                config_path (:obj:`str`): directory in which aws config file resides
        '''
        super().__init__(credential_path=aws_path, config_path=config_path, profile_name=profile_name)
        self.cache_dir = cache_dir
        self.profile_name = profile_name
        base_path_obj = Path(self.cache_dir)
        quilt3.session.AUTH_PATH = base_path_obj / 'auth.json'
        quilt3.session.CREDENTIALS_PATH = base_path_obj / 'credentials.json'
        quilt3.session.AUTH_PATH.touch()
        self.quilt_credentials_path = quilt3.session.CREDENTIALS_PATH

        self.aws_credentials_path = Path(aws_path) 

        dic = {'access_key': os.getenv('AWS_ACCESS_KEY_ID'),
               'secret_key': os.getenv('AWS_SECRET_ACCESS_KEY'),
               'token': None,
               'expiry_time': os.getenv('EXPIRY_TIME')}
        with open(str(self.quilt_credentials_path), 'w') as f:
            json.dump(dic, f)
        self.default_remote_registry = default_remote_registry
        quilt3.config(default_remote_registry=default_remote_registry)
        self.package = quilt3.Package()

    def bucket_obj(self, bucket_uri):
        ''' Create quilt3 bucket object

            Args:
                bucket_uri (:obj:`str`): quilt s3 bucket address

            Returns:
                (:obj:`quilt3.Bucket`): quilt3 bucket object
        '''
        return quilt3.Bucket(bucket_uri)

    def add_to_package(self, package_name, destination=None, source=None, meta=None):
        ''' Add file/directory to an existing package on S3 or build new package
            with name package_name
        
            Args:
                package_name (:obj:`str`): name of package to be modified
                source (:obj:`str`): source to be added to package,
                                    directories must end with '/'
                destination (:obj:`str` ): package to be manipulated,
                                        directories must end with '/'
                meta (:obj:`dict`): package meta

            Returns:

        '''
        suffix = '/'
        s = source
        m = meta
        d = destination
        try:
            p = quilt3.Package.install(package_name)
        except botocore.errorfactory.ClientError:
            p = self.package

        if s.endswith(suffix) != d.endswith(suffix): # when s and d do not share the same suffix
            return '{} and {} must have the same suffix. Operation stopped.'.format(d, s)

        if s.endswith(suffix):
            p.set_dir(d, s, meta=m)
        else:
            p.set(d, s, meta=m)
        p.build(name=package_name)

    def push_to_remote(self, package_name, registry=None, destination=None, message=None) -> str:
        ''' Push local package to remote registry

            Args:
                package_name (:obj:`str`): name of package in "username/packagename" format
                registry (:obj:`str`): S3 URI where package should be pushed
                destination (:obj:`str`): file landing destination in remote registry
                message (:obj:`str`): commit message
        '''
        try:
            p = quilt3.Package.browse(package_name)
        except botocore.errorfactory.ClientError:
            return 'No such package {} on s3'.format(package_name)
        except FileNotFoundError:
            self.package.build(name=package_name)
        except quilt3.util.QuiltException as e:
            return str(e)
        p.push(package_name, registry=registry, dest=destination, message=message)

    def delete_package(self, package_name, registry=None, delete_from_s3=True):
        """delete a local/remote package from local/remote registry
        
        Args:
            package_name (:obj:`str`): name of the package or name of file
            registry (:obj:`str`, optional): the registry from which the package will be removed. Defaults to None.
            delete_from_s3 (:obj:`bool`): whether to delete the underlying files in s3 bucket.
        """
        try:
            bucket = quilt3.Bucket(registry)
            bucket.delete_dir('.quilt/named_packages/' + package_name + '/')
        except quilt3.util.QuiltException as e:
            return str(e)
        if delete_from_s3:
            bucket.delete_dir(package_name + '/')

    def delete_obj(self, key):
        """delete an object from s3 bucket
        
        Args:
            key (:obj:`str`): key of the object in s3 bucket
        """
        try:
            bucket = quilt3.Bucket(self.default_remote_registry)
            bucket.delete(key)
        except botocore.errorfactory.ClientError:
            return 'No such object {} on s3'.format(key)        

    def build_from_external_bucket(self, package_name, package_dest, bucket_name, key, file_dir,
                                  bucket_credential=None, profile_name=None, meta=None,
                                  bucket_config=None, max_concurrency=10):
        ''' Build package with source from external (non-quilt) s3 buckets

            Args:
                package_name (:obj:`str`): local package name after being built
                package_dest (:obj:`str`): package(s) to be manipulated
                bucket_name (:obj:`str`): s3 bucket name
                key (:obj:`str`): the name of the key in s3 bucket to download from
                file_dir (:obj:`str`): the local path to which the file will be downloaded
                bucket_credential (:obj:`str`): directory in which credential for s3 bucket is stored
                profile_name (:obj:`str`): profile to be used for authentication
                meta (:obj:`dict`): meta information for package file
                bucket_credential (:obj:`str`): directory in which config for s3 bucket is stored
                max_concurrency (:obj:`int`): threads used for downloading
        '''
        settings = TransferConfig(max_concurrency=max_concurrency)
        if bucket_credential is None:
            bucket_credential = str(self.aws_credentials_path.expanduser())
        else:
            bucket_credential = str(Path(bucket_credential).expanduser())

        if bucket_config is None:
            bucket_config = str(self.aws_credentials_path.with_name('aws_config').expanduser())
        else:
            bucket_config = str(Path(bucket_config).expanduser())

        if profile_name is None:
            profile_name = self.profile_name

        s3 = s3_util.S3Util(profile_name=profile_name, credential_path=bucket_credential,
                            config_path=bucket_config)
                            
        file_name_path = Path(file_dir, key).expanduser()
        file_name = str(file_name_path)
        if package_dest.endswith('/'):
            file_name_path.mkdir(parents=True, exist_ok=True)
            s3.download_dir(key, bucket_name, local=file_dir)
        else:
            Path(file_name_path.parent).mkdir(parents=True, exist_ok=True)
            file_name_path.touch(exist_ok=True)        
            s3.client.download_file(bucket_name, key, file_name, Config=settings)
        
        if package_dest.endswith('/'):
            self.package.set_dir(package_dest, file_name, meta=meta)
        else:
            self.package.set(package_dest, file_name, meta=meta)
        return self.package.build(package_name)


import setuptools
try:
    import pkg_utils
except ImportError:
    import pip
    pip_version = tuple(pip.__version__.split('.'))
    if pip_version >= ('19', '3', '0'):
        import pip._internal.main as pip_main
    elif pip_version >= ('19', '0', '0'):
        import pip._internal as pip_main
    pip_main.main(['install', 'pkg_utils'])
    import pkg_utils
import os
import sys

name = 'karr_lab_aws_manager'
dirname = os.path.dirname(__file__)
package_data = {
    name: [
        'VERSION',
    ],
}

# get package metadata
md = pkg_utils.get_package_metadata(dirname, name, package_data_filename_patterns=package_data)

# install package
setuptools.setup(
    name=name,
    version=md.version,
    description='Package to manage aws services run by karr lab',
    long_description=md.long_description,
    url="https://github.com/KarrLab/" + name,
    download_url='https://github.com/KarrLab/' + name,
    author="Karr Lab",
    author_email="info@karrlab.org",
    license="MIT",
    keywords='aws boto3 datanator karrlab data',
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    package_data=md.package_data,
    install_requires=md.install_requires,
    extras_require=md.extras_require,
    tests_require=md.tests_require,
    dependency_links=md.dependency_links,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ],
    entry_points={
        'console_scripts': [
            '{} = {}.__main__:main'.format(name, name),
            '{} = {}.__main__:main'.format(name.replace('_', '-'), name),
            '{}{:d} = {}.__main__:main'.format(name, sys.version_info[0], name),
            '{}{:d} = {}.__main__:main'.format(name.replace('_', '-'), sys.version_info[0], name),
        ],
    },
)


from karr_lab_aws_manager.elasticsearch_kl import util
import requests


class QueryBuilder(util.EsUtil):

    def __init__(self, profile_name=None, credential_path=None,
                config_path=None, elastic_path=None,
                cache_dir=None, service_name='es', max_entries=float('inf'), verbose=False):
        ''' 
            Args:
                profile_name (:obj:`str`): AWS profile to use for authentication
                credential_path (:obj:`str`): directory for aws credentials file
                config_path (:obj:`str`): directory for aws config file
                elastic_path (:obj:`str`): directory for file containing aws elasticsearch service variables
                cache_dir (:obj:`str`): temp directory to store json for bulk upload
                service_name (:obj:`str`): aws service to be used
                max_entries (:obj:`int`): maximum number of operations
                verbose (:obj:`bool`): verbose messages
        '''
        super().__init__(profile_name=profile_name, credential_path=credential_path,
                config_path=config_path, elastic_path=elastic_path,
                cache_dir=cache_dir, service_name=service_name, max_entries=max_entries, verbose=verbose)


    def _set_options(self, query, option_key, option_value):
        ''' Builds query options for elasticsearch
            (https://opendistro.github.io/for-elasticsearch-docs/docs/elasticsearch/full-text/#options)

            Args:
                query_operation (:obj:`dict`): query body
                option_key (:obj:`str`): option name
                option_value (:obj:`str`) option value

            Returns:
                (:obj:`dict`): new query body
        '''
        query_operation = list(query['query'].keys())[0]
        query['query'][query_operation][option_key] = option_value
        return query
    
    def build_simple_query_string_body(self, query_message, **kwargs):
        ''' Builds query portion of the body in request body search
            (https://opendistro.github.io/for-elasticsearch-docs/docs/elasticsearch/full-text/#simple-query-string)

            Args:
                query_message (:obj:`str`): string to be queried for.
                
            Returns:
                (:obj:`dict`): query request body
        '''
        query_operation = 'simple_query_string'
        query = {'query': {query_operation: {'query': query_message}}}

        fields = kwargs.get('fields', ['*'])
        query = self._set_options(query, 'fields', fields)

        flags = kwargs.get('flags', 'ALL')
        query = self._set_options(query, 'flags', flags)

        fuzzy_transpositions = kwargs.get('fuzzy_transpositions', True)
        query = self._set_options(query, 'fuzzy_transpositions', fuzzy_transpositions)

        fuzzy_max_expansions = kwargs.get('fuzzy_max_expansions', 50)
        query = self._set_options(query, 'fuzzy_max_expansions', fuzzy_max_expansions)

        fuzzy_prefix_length = kwargs.get('fuzzy_prefix_length', 0)
        query = self._set_options(query, 'fuzzy_prefix_length', fuzzy_prefix_length)

        minimum_should_match = kwargs.get('minimum_should_match', 1)
        query = self._set_options(query, 'minimum_should_match', minimum_should_match)

        analyze_wildcard = kwargs.get('analyze_wildcard', True)
        query = self._set_options(query, 'analyze_wildcard', analyze_wildcard)

        lenient = kwargs.get('lenient', True)
        query = self._set_options(query, 'lenient', lenient)

        quote_field_suffix = kwargs.get('quote_field_suffix', "")
        query = self._set_options(query, 'quote_field_suffix', quote_field_suffix)

        auto_generate_synonyms_phrase_query = kwargs.get('auto_generate_synonyms_phrase_query', True)
        query = self._set_options(query, 'auto_generate_synonyms_phrase_query', auto_generate_synonyms_phrase_query)

        default_operator = kwargs.get('default_operator', 'OR')
        query = self._set_options(query, 'default_operator', default_operator)

        analyzer = kwargs.get('analyzer', 'standard')
        query = self._set_options(query, 'analyzer', analyzer)

        return query

    def build_bool_query_body(self, must=None, _filter=None, should=None, must_not=None,
                              minimum_should_match=0):
        """Building boolean query body
        (https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-bool-query.html)
        
        Args:
            must (:obj:`list` or :obj:`dict`, optional): Body for must. Defaults to None.
            _filter (:obj:`list` or :obj:`dict`, optional): Body for filter. Defaults to None.
            should (:obj:`list` or :obj:`dict`, optional): Body for should. Defaults to None.
            must_not (:obj:`list` or :obj:`dict`, optional): Body for must_not. Defaults to None.
            minimum_should_match (:obj:`int`): The number or percentage of should clauses returned documents must match. Defaults to 0.

        Return:
            (:obj:`dict`): boolean query body
        """
        query = {'query': {'bool': {"minimum_should_match": minimum_should_match}}}

        if must is not None:
            query['query']['bool']['must'] = must
        if _filter is not None:
            query['query']['bool']['filter'] = _filter
        if should is not None:
            query['query']['bool']['should'] = should
        if must_not is not None:
            query['query']['bool']['must_not'] = must_not

        return query


from elasticsearch import Elasticsearch, RequestsHttpConnection
from karr_lab_aws_manager.config import config
from karr_lab_aws_manager.elasticsearch_kl import index_setting_file
import requests
import json
from bson import ObjectId
import math
from pathlib import Path
from requests_aws4auth import AWS4Auth
import re
import datetime


class ComplexEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        elif isinstance(o, datetime.datetime):
            return o.__str__()
        return json.JSONEncoder.default(self, o)


class EsUtil(config.establishES):

    def __init__(self, profile_name=None, credential_path=None,
                config_path=None, elastic_path=None,
                cache_dir=None, service_name='es', max_entries=float('inf'), verbose=False):
        ''' 
            Args:
                profile_name (:obj:`str`): AWS profile to use for authentication
                credential_path (:obj:`str`): directory for aws credentials file
                config_path (:obj:`str`): directory for aws config file
                elastic_path (:obj:`str`): directory for file containing aws elasticsearch service variables
                cache_dir (:obj:`str`): temp directory to store json for bulk upload
                service_name (:obj:`str`): aws service to be used
        '''
        super().__init__(config_path=config_path, profile_name=profile_name, credential_path=credential_path,
                        elastic_path=elastic_path, service_name=service_name)
        self.verbose = verbose
        self.max_entries = max_entries
        self.cache_dir = cache_dir
        self.index_setting_manager = index_setting_file.IndexUtil()

    def unassigned_reason(self):
        """sends http request to get why a shard is unassigned

        Returns:
            (HTTPResponse): http response    
        """
        uri = self.es_endpoint + '/' + '_cat/shards?h=index,shard,prirep,state,unassigned.reason'
        r = requests.get(uri, auth=self.awsauth)
        return r

    def allocation_explain(self):
        """chooses the first unassigned shard that it finds and explains why it cannot be allocated to a node

        Returns:
            (HTTPResponse): http response    
        """
        uri = self.es_endpoint + '/_cluster/allocation/explain'
        r = requests.get(uri, auth=self.awsauth)
        return r

    def index_health_status(self):
        """shows the health status, number of documents, and disk usage for each index

        Returns:
            (HTTPResponse): http response    
        """
        uri = self.es_endpoint + '/_cat/indices?v'
        r = requests.get(uri, auth=self.awsauth)
        return r

    def get_index_mapping(self, index='.kibana_1'):
        """Get 
        
        Args:
            index (:obj:`str`, optional): Comma-separated list or wildcard expression of index names. Defaults to '.kibana_1'.

        Returns:
            (:obj:`requests.Response`)
        """
        uri = self.es_endpoint + '/' + index + '/_mapping'
        r = requests.get(uri, auth=self.awsauth)
        return r
    
    def index_settings(self, index, number_of_replicas, number_of_shards=1,
                      other_settings = {}, 
                      headers={ "Content-Type": "application/json" }):
        """Setting index's shard and replica number in es cluster
        
        Args:
            index (str): name of index to be set
            number_of_replicas (int): number of replica shards to be used for the index
            number_of_shards (int): number of primary shards contained in the es cluster
            other_settings (:obj:`dict`): other index settings.
            headers (dict): http request content header description

        Returns:
            (HTTPResponse): http response
        """
        url = self.es_endpoint + '/' + index + '/_settings'
        if number_of_shards == 1:
            body = {"index": {"number_of_replicas": number_of_replicas}}
        else:
            body = {"index": {"number_of_replicas": number_of_replicas,
                            "number_of_shards": number_of_shards}}
        
        if other_settings == {}:
            tmp = body['index']
        else:
            tmp = {**body['index'], **other_settings.get('index')}
        settings = {'index': tmp}
        r = requests.put(url, auth=self.awsauth, data=json.dumps(settings), headers=headers)
        return r

    def create_index(self, index, mappings=None, setting={"settings": {"number_of_shards": 1,
                    "number_of_replicas": 0}}, additional_settings=None):
        """Create index
            (https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html)

        Args:
            index (:obj:`str`): name of index
            setting (:obj:`dict`, optional): index settings. Defaults to {"settings": {"number_of_shards": 1}}.
            mappings (:obj:`dict`, optional): index mappings. Deafults to None.
            additional_settings (:obj:`dict`): additional settings. Defaults to None.
        """
        url = self.es_endpoint + '/' + index
        if mappings is not None:
            setting['mappings'] = mappings
        if additional_settings is not None:
            tot_setting = {**setting['settings'], **additional_settings}
            setting['settings'] = tot_setting
        r = requests.put(url, auth=self.awsauth, json=setting)
        return r

    def create_index_with_file(self, index, _file, num_shard=1, num_replica=0):
        """Create index with an index description file
            (https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html)

        Args:
            index (:obj:`str`): name of index
            _file (:obj:`dict`): index setting description.
            num_shard (:obj:`int`, optional): number of shards. Defaults to 1.
            num_replica (:obj:`int`, optional): number of replicas. Defaults to 0.

        Return:
            (:obj:`requests.Response`)
        """
        url = self.es_endpoint + '/' + index
        _file['settings']['number_of_shards'] = num_shard
        _file['settings']['number_of_replicas'] = num_replica
        print(_file)
        r = requests.put(url, auth=self.awsauth, json=_file)
        return r

    def migrate_index(self, old_index, new_index, headers={ "Content-Type": "application/json" },
                    number_of_shards=1, number_of_replicas=0):
        """Migrate old index to new index whilst changing shard and replica setting
        
        Args:
            old_index (:obj:`str`): name of the old index.
            new_index (:obj:`str`): name of the new index
            headers (:obj:`HTTP.header`, optional): header. Defaults to { "Content-Type": "application/json" }.
            number_of_shards (:obj:`int`, optional): number of shards for the index. Defaults to 1.
            number_of_replicas (:obj:`int`, optional): number of replicas for the index. Defaults to 1.

        Return:
            (:obj:`list` of :obj:`requests.Response`)
        """
        template = {"index_patterns": [new_index], "settings": {"number_of_shards": number_of_shards,
                                                                "number_of_replicas": number_of_replicas}}
        template_url = self.es_endpoint + '/_template/template_1'
        reindex = {"source": {"index": old_index}, "dest": {"index": new_index}}
        reindex_url = self.es_endpoint + '/_reindex'
        count_url = self.es_endpoint + '/_cat/count/'
        old_index_url = self.es_endpoint + '/' + old_index
        r_template = requests.put(template_url, auth=self.awsauth, json=template)
        r_reindex = requests.post(reindex_url, auth=self.awsauth, json=reindex)
        count_old = requests.get(count_url+old_index, auth=self.awsauth).content.decode('utf-8')
        count_new = requests.get(count_url+new_index, auth=self.awsauth).content.decode('utf-8')
        if count_old.split(' ')[:-1] == count_new.split(' ')[:-1]:
            r = requests.delete(old_index_url, auth=self.awsauth)
        else:
            r = [count_old, count_new]
        return [r_template, r_reindex, r]     

    def put_mapping(self, index, body):
        """Put index mapping to exisiting index.
        (https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-put-mapping.html)
        
        Args:
            index (:obj:`str`): mapping for the index.
            body (:obj:`dict`): mapping description.

        Return:
            (:obj:`requests.Response`)
        """
        url = self.es_endpoint + '/' + index + '/' + '_mapping'
        r = requests.put(url, auth=self.awsauth, json=body)
        return r

    def enable_fielddata(self, index, _type, field):
        """Enable fielddata for type fields
        
        Args:
            index (:obj:`str`): Index in which the operation will be done
            _type (:obj:`str`): Existing mapping for field.
            field (:obj:`str`): name of the field.
        """
        url = self.es_endpoint + '/' + index + '/_mapping'
        body = {
                "properties": {
                    field: { 
                    "type": _type,
                    "fielddata": True
                    }
                }}
        r = requests.put(url, auth=self.awsauth, json=body)
        return r

    def build_es(self, suffix=None):
        ''' build es query object

            Args:
                suffix (:obj:`str`): string trailing es endpoint

            Returns:
                (:obj:`Elasticsearch`): Elasticsearch object
        '''
        if suffix is None:
            uri = self.es_endpoint.split('https://')[1]
        else:
            uri = self.es_endpoint.split('https://')[1] + suffix 
        
        es = Elasticsearch(
            hosts = [{'host': uri, 'port': 443}],
            http_auth = self.awsauth,
            use_ssl = True,
            verify_certs = True,
            connection_class = RequestsHttpConnection
        )
        return es

    def add_field_to_index(self, index, field, value):
        """Add a field of value to all documents in index
        
        Args:
            index (:obj:`str`): name of index.
            field (:obj:`str`): name of field.
            value (:obj:`Obj`): value of field.

        Return:
            (:obj:`HTTPResponse`): elasticsearch update status description.
        """
        url = self.es_endpoint + '/' + index + '/' + '_update_by_query'
        if isinstance(value, (int, float, complex)) and not isinstance(value, bool):
            val = str(value)
        else:
            val = "\""+value+"\""
        script = "ctx._source.{} = {}".format(field, val)
        body = {
                    "query": {
                        "match_all": {}
                    },
                    "script": {"inline": script}
                }
        r = requests.post(url, auth=self.awsauth, json=body)
        return r

    def make_action_and_metadata(self, index, _id):
        ''' Make action_and_metadata obj for bulk loading
            e.g. { "index": { "_index" : "index", "_id" : "id" } }

            Args:
                index (:obj:`str`): name of index on ES
                _id (:obj:`str`):  unique id for document

            Returns:
                (:obj:`dict`): metadata that conforms to ES bulk load requirement
        '''
        action_and_metadata = {'index': { "_index" : index, "_id" : _id }}
        return action_and_metadata

    def delete_index(self, index, _id=None):
        ''' Delete elasticsearch index

            Args:
                index (:obj:`str`): name of index in es
                _id (:obj:`int`): id of the doc in index (optional)
        '''
        if _id is None:
            url = self.es_endpoint + '/' + index
        else:
            url = self.es_endpoint + '/' + index + '/_doc/' + _id
        r = requests.delete(url, auth=self.awsauth)
        return r.status_code

    def data_to_es_bulk(self, cursor, index='test', count=None, bulk_size=100, _id='uniprot_id',
                        headers={ "Content-Type": "application/json" }):
        ''' Load data into elasticsearch service

            Args:
                count (:obj:`int`): cursor size
                cursor (:obj:`pymongo.Cursor` or :obj:`iter`): documents to be PUT/POST to es
                index (:obj:`str`): name of unique key to be used as index for es
                bulk_size (:obj:`int`): number of documents in one PUT
                headers (:obj:`dict`): http header
                _id (:obj:`str`): key in mogno collection for identification

            Returns:
                (:obj:`set`): set of status codes
        '''
        def gen_bulk_file(_iid, bulk_file):
            action_and_metadata = self.make_action_and_metadata(index, _iid)
            bulk_file += json.dumps(action_and_metadata, cls=ComplexEncoder) + '\n'
            if doc.get('_id') is not None:
                doc['id'] = doc.pop('_id')
            bulk_file += json.dumps(doc, cls=ComplexEncoder) + '\n'  
            return bulk_file

        def mod_cursor(cursor):
            pathlist = Path(cursor).expanduser().glob('**/*.json')
            for path in pathlist:
                with path.open() as f:
                    yield json.load(f)     

        if isinstance(cursor, str):
            count = len(list(Path(cursor).expanduser().glob('**/*.json'))) 
            cursor = mod_cursor(cursor)            
        elif isinstance(cursor, list):
            count = len(cursor) 
        else:
            cursor = cursor       

        url = self.es_endpoint + '/_bulk'
        status_code = {201}
        bulk_file = ''
        tot_rounds = math.ceil(count/bulk_size)        

        for i, doc in enumerate(cursor):
            if i == self.max_entries:
                break
            if self.verbose and i % bulk_size == 0:
                print("Processing bulk {} out of {} ...".format(math.floor(i/bulk_size)+1, tot_rounds))
               
            if i == count - 1:  # last entry
                bulk_file = gen_bulk_file(doc[_id], bulk_file)
                r = requests.post(url, auth=self.awsauth, data=bulk_file, headers=headers)
                status_code.add(r.status_code)
                return r
            elif i % bulk_size != 0 or i == 0: #  bulk_size*(n-1) + 1 --> bulk_size*n - 1
                bulk_file = gen_bulk_file(doc[_id], bulk_file)
            else:               # bulk_size * n
                r = requests.post(url, auth=self.awsauth, data=bulk_file, headers=headers)
                status_code.add(r.status_code)
                bulk_file = gen_bulk_file(doc[_id], '') # reset bulk_file
                
    def data_to_es_single(self, count, cursor, index, _id='uniprot_id',
                          headers={ "Content-Type": "application/json" }):
        ''' Load data into elasticsearch service

            Args:
                count (:obj:`int`): cursor size
                cursor (:obj:`pymongo.Cursor` or :obj:`iter`): documents to be PUT to es
                index (:obj:`str`): name of unique key to be used as index for es
                es_endpoint (:obj:`str`): elasticsearch endpoint
                headers (:obj:`dict`): http header information
                _id (:obj:`str`): key in mongo collection for identification
                
            Returns:
                (:obj:`set`): set of status codes
        '''
        url_root = self.es_endpoint + '/' + index + '/_doc/'
        status_code = {201}
        for i, doc in enumerate(cursor):
            if i == self.max_entries:
                break
            if i % 20 == 0 and self.verbose:
                print("Processing doc {} out of {}...".format(i, min(count, self.max_entries)))
            doc = json.loads(json.dumps(doc, cls=ComplexEncoder))
            url = url_root + doc[_id]
            r = requests.post(url, auth=self.awsauth, json=doc, headers=headers)
            status_code.add(r.status_code)
        return status_code

def main():
    manager = EsUtil(profile_name='es-poweruser', credential_path='~/.wc/third_party/aws_credentials',
                config_path='~/.wc/third_party/aws_config', elastic_path='~/.wc/third_party/elasticsearch.ini',
                service_name='es', max_entries=float('inf'), verbose=True)
    r = manager.enable_fielddata('protein', 'text', 'ko_number')
    print(r.text)

if __name__ == "__main__":
    main()